# Copyright (c) 2025, Giampaolo Rodola. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import functools
import sys

import pytest

import psleak

# speed things up
psleak.MemoryLeakTestCase.times = 50
psleak.MemoryLeakTestCase.retries = 5
psleak.MemoryLeakTestCase.warmup_times = 2

# how many times retry_on_failure() decorator will retry
NO_RETRIES = 10


def retry_on_failure(retries=NO_RETRIES):
    """Decorator which runs a test function and retries N times before
    giving up and failing.
    """

    def decorator(test_method):
        @functools.wraps(test_method)
        def wrapper(self, *args, **kwargs):
            err = None
            for attempt in range(retries):
                try:
                    return test_method(self, *args, **kwargs)
                except (  # noqa: PERF203
                    AssertionError,
                    pytest.fail.Exception,
                ) as _:
                    err = _
                    prefix = "\n" if attempt == 0 else ""
                    short_err = str(err).split("\n")[0]
                    print(  # noqa: T201
                        f"{prefix}{short_err}, retrying"
                        f" {attempt + 1}/{retries} ...",
                        file=sys.stderr,
                    )
                    if hasattr(self, "tearDown"):
                        self.tearDown()
                    if hasattr(self, "teardown_method"):
                        self.teardown_method()
                    if hasattr(self, "setUp"):
                        self.setUp()
                    if hasattr(self, "setup_method"):
                        self.setup_method()

            raise err

        return wrapper

    assert retries > 1, retries
    return decorator



# Copyright (c) 2025, Giampaolo Rodola. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""Test framework to detect memory and resource leaks in Python C
extensions.
"""

import collections
import functools
import gc
import logging
import os
import sys
import threading
import types
import unittest
import warnings
from dataclasses import dataclass

import psutil
from psutil import POSIX
from psutil import WINDOWS
from psutil._common import print_color

thisproc = psutil.Process()


# --- exceptions


class Error(AssertionError):
    """Base class for all psleak exceptions."""


class UnclosedResourceError(Error):
    """Base class for errors raised when some resource created during a
    function call is left unclosed or unfreed afterward.
    """

    resource_name = "resource"
    verb = "unclosed"

    def __init__(self, count, fun_name, extras=None):
        self.count = count
        self.fun_name = fun_name
        self.extras = extras
        name = self.resource_name
        name += "s" if count > 1 else ""  # pluralize
        msg = (
            f"detected {count} {self.verb} {name} after calling {fun_name!r} 1"
            " time"
        )
        if extras:
            msg += ":" + "".join(f"\n* {extra!r}" for extra in extras)
        super().__init__(msg)


class UnclosedFdError(UnclosedResourceError):
    """Raised when an unclosed file descriptor is detected after
    calling function once. Used to detect forgotten close(). UNIX only.
    """

    resource_name = "file descriptor"


class UnclosedHandleError(UnclosedResourceError):
    """Raised when an unclosed handle is detected after calling
    function once. Used to detect forgotten CloseHandle().
    Windows only.
    """

    resource_name = "handle"


class UnclosedHeapCreateError(UnclosedResourceError):
    """Raised when test detects HeapCreate() without a corresponding
    HeapDestroy() after calling function once. Windows only.
    """

    resource_name = "HeapCreate() call"


class UnclosedNativeThreadError(UnclosedResourceError):
    """Raised when a native C thread created outside Python is running
    after calling function once. Detects pthread_create() without
    a corresponding pthread_join().
    """

    resource_name = "native C thread"


class UnclosedPythonThreadError(UnclosedResourceError):
    """Raised when a Python thread is running after calling function
    once. This indicates that a `threading.Thread` was start()ed but not
    properly join()ed or stopped.
    """

    resource_name = "Python thread"


class UncollectableGarbageError(UnclosedResourceError):
    """Raised when objects with __del__ are left in gc.garbage after a call."""

    resource_name = "GC object"
    verb = "uncollectable"


class MemoryLeakError(Error):
    """Raised when a memory leak is detected after calling function
    many times. Aims to detect:

    - `malloc()` without a corresponding `free()`
    - `mmap()` without `munmap()`
    - `HeapAlloc()` without `HeapFree()` (Windows)
    - `VirtualAlloc()` without `VirtualFree()` (Windows)
    """


# --- utils


def format_run_line(idx, diffs, times):
    parts = [f"{k}={'+' + str(v):<6}" for k, v in diffs.items() if v > 0]
    metrics = " | ".join(parts)
    avg = "0B"
    if parts:
        first_key = next(k for k, v in diffs.items() if v > 0)
        avg = str(diffs[first_key] // times)
    s = f"Run #{idx:>2}: {metrics:<50} (calls={times:>4}, avg/call=+{avg})"
    if idx == 1:
        s = "\n" + s
    return s


def qualname(obj):
    """Return a human-readable qualified name for a function, method or
    class.
    """
    return getattr(obj, "__qualname__", getattr(obj, "__name__", str(obj)))


def assert_isinstance(name, obj, types):
    if not isinstance(obj, types):
        if isinstance(types, tuple):
            exp = " or ".join(t.__name__ for t in types)
        else:
            exp = types.__name__
        msg = f"{name!r} must be instance of {exp} (got {obj!r})"
        raise TypeError(msg)


# --- GC debugger


class GCDebugger:
    """Detects objects that cannot be automatically garbage collected
    because they form reference cycles or define a finalizer method
    (__del__).

    Detection is performed using a context manager that temporarily
    enables gc.DEBUG_SAVEALL and tracks objects remaining in
    gc.garbage.
    """

    # Objects that are temporarily part of a cycle but are expected to
    # disappear once the cycle is broken.
    TRANSIENT_TYPES = (
        types.FrameType,
        types.TracebackType,
        type(threading.current_thread()),
        BaseException,  # ignore all exception instances
    )

    # Value-like objects that do not hold references to other Python
    # objects tracked by the GC and therefore cannot participate in
    # reference cycles.
    SCALAR_TYPES = (
        int,
        float,
        bool,
        str,
        bytes,
        bytearray,
        complex,
        type(None),
    )

    def __enter__(self):
        self._old_debug = gc.get_debug()
        gc.set_debug(gc.DEBUG_SAVEALL)
        gc.collect()
        self.before = list(gc.garbage)
        self.after = []
        gc.garbage.clear()
        return self

    def __exit__(self, *a, **k):
        gc.collect()
        self.after = list(gc.garbage)
        gc.garbage.clear()
        gc.set_debug(self._old_debug)

    def is_transient(self, obj, _seen=None):
        if _seen is None:
            _seen = set()

        oid = id(obj)
        if oid in _seen:
            return True

        _seen.add(oid)

        if isinstance(obj, self.TRANSIENT_TYPES):
            return True

        if isinstance(obj, self.SCALAR_TYPES):
            return True

        if isinstance(obj, (list, tuple, set, frozenset)):
            for o in obj:  # noqa: SIM110
                if not self.is_transient(o, _seen):
                    return False
            return True

        if isinstance(obj, dict):
            for k, v in obj.items():
                if not self.is_transient(k, _seen):
                    return False
                if not self.is_transient(v, _seen):
                    return False
            return True

        return False

    def leaked_objects(self):
        leaked = []
        for obj in self.after:
            if obj in self.before:
                continue
            if self.is_transient(obj):
                continue
            leaked.append(obj)
        return leaked

    def check(self, fun):
        leaked = self.leaked_objects()
        if leaked:
            raise UncollectableGarbageError(
                len(leaked), qualname(fun), extras=leaked
            )


# --- checkers config


@dataclass(frozen=True)
class Checkers:
    """Configuration object controlling which leak checkers are enabled."""

    # C stuff
    memory: bool = True
    fds: bool = True
    handles: bool = True
    c_threads: bool = True
    # Python stuff
    py_threads: bool = True
    gcgarbage: bool = True

    @classmethod
    def _validate(cls, check_names):
        """Validate checker names and return set of all fields."""
        all_fields = set(cls.__annotations__.keys())
        invalid = set(check_names) - all_fields
        if invalid:
            msg = f"invalid checker names: {', '.join(invalid)}"
            raise ValueError(msg)
        return all_fields

    @classmethod
    def only(cls, *checks):
        """Return a config object with only the specified checkers enabled."""
        all_fields = cls._validate(checks)
        kwargs = {f: f in checks for f in all_fields}
        return cls(**kwargs)

    @classmethod
    def exclude(cls, *checks):
        """Return a config object with the specified checkers disabled."""
        all_fields = cls._validate(checks)
        kwargs = {f: f not in checks for f in all_fields}
        return cls(**kwargs)


# ---

_warnings_emitted = False


def _emit_warnings():
    global _warnings_emitted  # noqa: PLW0603

    if _warnings_emitted:
        return

    # PYTHONMALLOC check
    if not os.environ.get("PYTHONMALLOC", "").startswith(
        ("malloc", "mimalloc")
    ):
        msg = (
            "PYTHONMALLOC=malloc environment variable was not set; memory leak"
            " detection may be less reliable"
        )
        warnings.warn(msg, RuntimeWarning, stacklevel=2)

    # PYTHONUNBUFFERED check
    if os.environ.get("PYTHONUNBUFFERED") != "1":
        msg = (
            "PYTHONUNBUFFERED=1 environment variable was not set; memory leak"
            " detection may be less reliable"
        )
        warnings.warn(msg, RuntimeWarning, stacklevel=2)

    # pytest-xdist check
    if "PYTEST_XDIST_WORKER" in os.environ:
        msg = (
            "memory leak detection is unreliable when running tests in"
            " parallel via pytest-xdist"
        )
        warnings.warn(msg, RuntimeWarning, stacklevel=2)

    # background threads
    if threading.active_count() > 1:
        msg = (
            "active Python threads exist before test; memory/thread counts may"
            f" be unreliable: {threading.enumerate()}"
        )
        warnings.warn(msg, RuntimeWarning, stacklevel=2)

    # heap info
    if not hasattr(psutil, "heap_info"):  # SunOS, OpenBSD
        msg = (
            "psutil.heap_info() not available on this platform; memory leak"
            " detection is less reliable"
        )
        warnings.warn(msg, RuntimeWarning, stacklevel=2)

    _warnings_emitted = True


class MemoryLeakTestCase(unittest.TestCase):
    # Warm-up calls before starting measurement.
    warmup_times = 10
    # Number of times to call the tested function in each iteration.
    times = 200
    # Maximum retries if memory keeps growing.
    retries = 10
    # Allowed memory growth (in bytes or per-metric) before it is
    # considered a leak.
    tolerance = 0
    # Optional callable to free caches before starting measurement.
    trim_callback = None
    # Config object which tells which checkers to run.
    checkers = Checkers()
    # 0 = no messages; 1 = print diagnostics when memory increases.
    verbosity = 1

    __doc__ = __doc__

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_fds = self._get_fds()

    @classmethod
    def setUpClass(cls):
        cls._psutil_debug_orig = bool(os.getenv("PSUTIL_DEBUG"))
        psutil._set_debug(False)  # avoid spamming to stderr

    @classmethod
    def tearDownClass(cls):
        psutil._set_debug(cls._psutil_debug_orig)

    def _log(self, msg, level):
        if level <= self.verbosity:
            if WINDOWS:
                # On Windows we use ctypes to add colors. Avoid that to
                # not interfere with memory observations.
                print(msg)  # noqa: T201
            else:
                print_color(msg, color="yellow")
            # Force flush to not interfere with memory observations.
            sys.stdout.flush()

    def _trim_mem(self):
        """Release unused memory. Aims to stabilize memory measurements."""
        if self._trim_callback is not None:
            self._trim_callback()

        # flush standard streams
        for stream in (sys.stdout, sys.stderr, sys.__stdout__, sys.__stderr__):
            stream.flush()

        # flush logging handlers
        for handler in logging.root.handlers:
            handler.flush()

        # Full garbage collection. Note: cPython does it 3 times, but
        # it seems more historical churn.
        # https://github.com/giampaolo/cpython/blob/2e27da18952/Lib/test/support/__init__.py
        gc.collect()
        if gc.garbage:
            msg = f"GC garbage is not empty: {gc.garbage}"
            raise AssertionError(msg)

        if hasattr(sys, "_clear_internal_caches"):  # python 3.13
            sys._clear_internal_caches()
        elif hasattr(sys, "_clear_type_cache"):
            sys._clear_type_cache()

        # release free heap memory back to the OS
        if hasattr(psutil, "heap_trim"):
            psutil.heap_trim()

    def _warmup(self, fun, warmup_times):
        for _ in range(warmup_times):
            self.call(fun)

    # --- getters

    def _get_fds(self):
        """Return regular files and socket connections opened by
        process. Other FD types (e.g. pipes or dirs) won't be listed.
        """
        ls = []
        try:
            ls.extend(thisproc.open_files())
        except psutil.Error:
            pass
        try:
            ls.extend(thisproc.net_connections(kind="all"))
        except psutil.Error:
            pass
        return ls

    def _get_counters(self, checkers):
        # order matters
        d = {}
        if checkers.py_threads:
            d["py_threads"] = (
                threading.active_count(),
                threading.enumerate(),
            )
        if POSIX and checkers.fds:
            d["num_fds"] = (thisproc.num_fds(), self._cached_fds)
        if WINDOWS and checkers.handles:
            d["num_handles"] = (thisproc.num_handles(), self._cached_fds)
        if checkers.c_threads:
            d["c_threads"] = (thisproc.num_threads(), thisproc.threads())
        if WINDOWS and checkers.memory:
            d["heap_count"] = (psutil.heap_info().heap_count, [])
        return d

    def _get_mem(self):
        mem = thisproc.memory_full_info()
        heap_used = mmap_used = 0
        if hasattr(psutil, "heap_info"):
            heap = psutil.heap_info()
            heap_used = heap.heap_used
            mmap_used = heap.mmap_used
        return {
            "heap": heap_used,
            "mmap": mmap_used,
            "uss": getattr(mem, "uss", 0),
            "rss": mem.rss,
            "vms": mem.vms,
        }

    # --- checkers

    def _check_counters(self, fun, checkers):
        before = self._get_counters(checkers)
        self.call(fun)
        after = self._get_counters(checkers)

        for what, (count_before, extras_before) in before.items():
            count_after = after[what][0]
            extras_after = after[what][1]
            diff = count_after - count_before

            if diff < 0:
                msg = (
                    f"WARNING: {what!r} decreased by {abs(diff)} after calling"
                    f" {qualname(fun)!r} once"
                )
                self._log(msg, 0)

            elif diff > 0:
                if what in {"num_fds", "num_handles"}:
                    # fetch fds and update cache only in case of failure
                    extras_after = self._cached_fds = self._get_fds()

                extras = set(extras_after) - set(extras_before)
                mapping = {
                    "num_fds": UnclosedFdError,
                    "num_handles": UnclosedHandleError,
                    "heap_count": UnclosedHeapCreateError,
                    "py_threads": UnclosedPythonThreadError,
                    "c_threads": UnclosedNativeThreadError,
                }
                exc = mapping.get(what)
                if exc is None:
                    raise ValueError(what)
                raise exc(diff, qualname(fun), extras=extras)

    def _call_ntimes(self, fun, times):
        """Get memory samples before and after calling fun repeatedly,
        and return the diffs as a dict.
        """
        self._trim_mem()
        mem1 = self._get_mem()

        for _ in range(times):
            self.call(fun)

        self._trim_mem()
        mem2 = self._get_mem()

        diffs = {k: mem2[k] - mem1[k] for k in mem1}
        return diffs

    def _check_mem(self, fun, times, retries, tolerance):
        prev = {}
        messages = []
        if isinstance(tolerance, dict):
            tolerances = tolerance
        else:
            t = 0 if tolerance is None else tolerance
            tolerances = dict.fromkeys(self._get_mem(), t)

        increase = int(times / 2)  # 50%
        for idx in range(1, retries + 1):
            diffs = self._call_ntimes(fun, times)
            leaks = {k: v for k, v in diffs.items() if v > 0}

            if leaks:
                line = format_run_line(idx, leaks, times)
                messages.append(line)
                self._log(line, 1)

            # stable means:
            # * any growth is within tolerance, OR
            # * growth has stopped (no increase vs prev)
            stable = all(
                diffs[k] <= tolerances.get(k, 0) or diffs[k] <= prev.get(k, 0)
                for k in diffs
            )

            if stable:
                if idx > 1 and leaks:
                    self._log(
                        "Memory stabilized (no further growth detected)", 1
                    )
                return

            prev = diffs
            times += increase

        msg = f"memory kept increasing after {retries} runs" + "\n".join(
            messages
        )
        raise MemoryLeakError(msg)

    def _validate_opts(
        self, warmup_times, times, retries, tolerance, trim_callback
    ):
        assert_isinstance("warmup_times", warmup_times, int)
        assert_isinstance("times", times, int)
        assert_isinstance("retries", retries, int)
        assert_isinstance("tolerance", tolerance, (int, dict))
        if trim_callback is not None:
            assert_isinstance(
                "trim_callback", trim_callback, collections.abc.Callable
            )

        if warmup_times < 0:
            msg = f"warmup_times must be >= 0 (got {warmup_times})"
            raise ValueError(msg)
        if times < 1:
            msg = f"times must be >= 1 (got {times})"
            raise ValueError(msg)
        if retries < 0:
            msg = f"retries must be >= 0 (got {retries})"
            raise ValueError(msg)
        if tolerance is not None:
            if isinstance(tolerance, int):
                if tolerance < 0:
                    msg = f"tolerance must be >= 0 (got {tolerance!r})"
                    raise ValueError(msg)
            else:
                mem_keys = self._get_mem().keys()
                for k, v in tolerance.items():
                    if k not in mem_keys:
                        msg = f"invalid tolerance key {k!r}"
                        raise ValueError(msg)
                    if v < 0:
                        msg = f"{k!r} tolerance must be >= 0 (got {v})"
                        raise ValueError(msg)

    # ---

    def call(self, fun):
        return fun()

    def execute(
        self,
        fun,
        *args,
        warmup_times=None,
        times=None,
        retries=None,
        tolerance=None,
        trim_callback=None,
        checkers=None,
    ):
        """Run a full leak test on a callable. If specified, the
        optional arguments override the class attributes with the same
        name.
        """
        warmup_times = (
            warmup_times if warmup_times is not None else self.warmup_times
        )
        times = times if times is not None else self.times
        retries = retries if retries is not None else self.retries
        tolerance = tolerance if tolerance is not None else self.tolerance
        checkers = checkers if checkers is not None else self.checkers
        trim_callback = (
            trim_callback if trim_callback is not None else self.trim_callback
        )

        self._validate_opts(
            warmup_times, times, retries, tolerance, trim_callback
        )

        _emit_warnings()

        if args:
            fun = functools.partial(fun, *args)

        self._trim_callback = trim_callback

        # run check counters
        if checkers.gcgarbage:
            with GCDebugger() as gcdbg:
                self._check_counters(fun, checkers)
            gcdbg.check(fun)
        else:
            self._check_counters(fun, checkers)

        # run memory checks
        if checkers.memory:
            self._warmup(fun, warmup_times)
            self._check_mem(
                fun, times=times, retries=retries, tolerance=tolerance
            )

    def execute_w_exc(self, exc, fun, *args, **kwargs):
        """Run MemoryLeakTestCase.execute() expecting fun() to raise
        exc on every call.

        The exception is caught so resource and memory checks can run
        normally. If `fun()` does not raise `exc` on any call, the
        test fails.
        """

        def call():
            try:
                self.call(fun)
            except exc:
                pass
            else:
                return self.fail(f"{qualname(fun)!r} did not raise {exc}")

        if args:
            fun = functools.partial(fun, *args)

        self.execute(call, **kwargs)



# Copyright (c) 2025, Giampaolo Rodola. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import contextlib
import io
import os
import socket
import threading
import time
import warnings
from unittest import mock

import pytest
from psutil import NETBSD
from psutil import POSIX
from psutil import WINDOWS

import psleak
from psleak import Checkers
from psleak import MemoryLeakError
from psleak import MemoryLeakTestCase
from psleak import UnclosedFdError
from psleak import UnclosedHandleError
from psleak import _emit_warnings

from . import retry_on_failure


class TestMisc(MemoryLeakTestCase):
    def test_success(self):
        def foo():
            return 1 + 1

        self.execute(foo)

    @retry_on_failure()
    def test_leak_mem(self):
        ls = []

        def fun(ls=ls):
            ls.append("x" * 248 * 1024)

        try:
            # will consume around 60M in total
            with pytest.raises(MemoryLeakError):
                with contextlib.redirect_stdout(
                    io.StringIO()
                ), contextlib.redirect_stderr(io.StringIO()):
                    self.execute(fun, times=100)
        finally:
            del ls

    def test_unclosed_file(self):
        def fun():
            f = open(__file__)  # noqa: SIM115
            self.addCleanup(f.close)
            box.append(f)  # prevent auto-gc

        box = []
        with pytest.raises(
            UnclosedFdError if POSIX else UnclosedHandleError
        ) as cm:
            self.execute(fun)
        if not NETBSD:
            assert len(cm.value.extras) == 1
            extra = cm.value.extras.pop()
            assert os.path.normpath(extra.path) == os.path.normpath(__file__)

    def test_unclosed_socket(self):
        def fun():
            sock = socket.socket()
            sock.bind(("", 0))
            sock.listen(5)
            self.addCleanup(sock.close)
            box.append(sock)  # prevent auto-gc

        box = []
        with pytest.raises(
            UnclosedFdError if POSIX else UnclosedHandleError
        ) as cm:
            self.execute(fun)
        assert len(cm.value.extras) == 1
        assert "SOCK_STREAM" in str(cm)

    @pytest.mark.skipif(not POSIX, reason="POSIX only")
    def test_unclosed_fd(self):
        def fun():
            fd = os.open("/dev/null", os.O_RDONLY)
            self.addCleanup(os.close, fd)
            box.append(fd)  # prevent auto-gc

        box = []

        with pytest.raises(UnclosedFdError) as cm:
            self.execute(fun)
        assert cm.value.count == 1
        assert "1 unclosed file descriptor" in str(cm)

    @pytest.mark.skipif(not WINDOWS, reason="WINDOWS only")
    def test_unclosed_handles(self):
        import _winapi  # noqa: PLC0415

        def fun():
            handle = _winapi.OpenProcess(
                _winapi.PROCESS_ALL_ACCESS, False, os.getpid()
            )
            self.addCleanup(_winapi.CloseHandle, handle)

        with pytest.raises(UnclosedHandleError):
            self.execute(fun)

    def test_tolerance(self):
        def fun():
            ls.append("x" * 24 * 1024)

        ls = []
        times = 100
        self.execute(
            fun, times=times, warmup_times=0, tolerance=200 * 1024 * 1024
        )

    def test_tolerance_dict(self):
        ls = []

        def fun():
            ls.append("x" * 24 * 1024)

        n = 200 * 1024 * 1024

        # integer tolerance large enough
        self.execute(fun, times=100, warmup_times=0, tolerance=n)

        # None tolerance (same as 0)
        ls.clear()
        with pytest.raises(MemoryLeakError):
            self.execute(fun, warmup_times=0, tolerance=None)

        # dict full tolerance
        ls.clear()
        tol = {"rss": n, "heap": n, "mmap": n, "uss": n, "vms": n}
        self.execute(fun, warmup_times=0, tolerance=tol)

        # dict full tolerance except some
        ls.clear()
        tol = {"rss": 0, "heap": 0, "mmap": n, "uss": 0, "vms": 0}
        with pytest.raises(MemoryLeakError):
            self.execute(fun, warmup_times=0, tolerance=tol)

    def test_tolerance_errors(self):
        def fun():
            pass

        # negative integer
        with pytest.raises(ValueError, match="tolerance must be >= 0"):
            self.execute(fun, times=1, tolerance=-1)
        # invalid dict key
        with pytest.raises(
            ValueError, match="invalid tolerance key 'nonexistent'"
        ):
            self.execute(fun, times=1, tolerance={"nonexistent": 10})

        # invalid tolerance type
        with pytest.raises(TypeError, match="must be instance of"):
            self.execute(fun, times=1, tolerance="invalid")

    def test_execute_args_validation(self):
        def fun():
            pass

        # type errors
        with pytest.raises(TypeError):
            self.execute(fun, warmup_times="10")
        with pytest.raises(TypeError):
            self.execute(fun, times="100")
        with pytest.raises(TypeError):
            self.execute(fun, retries="100")
        with pytest.raises(TypeError):
            self.execute(fun, tolerance="bad")
        with pytest.raises(TypeError):
            self.execute(fun, trim_callback=123)

        # value errors
        with pytest.raises(ValueError):  # noqa: PT011
            self.execute(fun, warmup_times=-1)
        with pytest.raises(ValueError):  # noqa: PT011
            self.execute(fun, times=0)
        with pytest.raises(ValueError):  # noqa: PT011
            self.execute(fun, retries=-1)
        with pytest.raises(ValueError):  # noqa: PT011
            self.execute(fun, tolerance=-1)
        with pytest.raises(ValueError):  # noqa: PT011
            self.execute(fun, tolerance={"invalid": 1})
        with pytest.raises(ValueError):  # noqa: PT011
            self.execute(fun, tolerance={"rss": -1})

        with pytest.raises(ValueError, match="times must be"):
            self.execute(lambda: 0, times=0)
        with pytest.raises(ValueError, match="times must be"):
            self.execute(lambda: 0, times=-1)
        with pytest.raises(ValueError, match="warmup_times"):
            self.execute(lambda: 0, warmup_times=-1)
        with pytest.raises(ValueError, match="tolerance"):
            self.execute(lambda: 0, tolerance=-1)
        with pytest.raises(ValueError, match="retries"):
            self.execute(lambda: 0, retries=-1)

    def test_execute_w_exc(self):
        def fun_1():
            1 / 0  # noqa: B018

        self.execute_w_exc(ZeroDivisionError, fun_1)

        with pytest.raises(ZeroDivisionError):
            self.execute_w_exc(OSError, fun_1)

        def fun_2(a):
            pass

        with pytest.raises(AssertionError, match="did not raise"):
            self.execute_w_exc(ZeroDivisionError, fun_2, 1)

    def test_trim_callback(self):
        called = []

        def cleanup():
            called.append(True)

        def fun():
            pass

        class MyTest(MemoryLeakTestCase):
            pass

        tc = MyTest()
        tc.execute(fun, trim_callback=cleanup)
        assert called


class TestCheckers:

    def test_default_values(self):
        checkers = Checkers()
        assert checkers.fds
        assert checkers.handles
        assert checkers.py_threads
        assert checkers.c_threads
        assert checkers.memory
        assert checkers.gcgarbage

    def test_only(self):
        checkers = Checkers.only("fds", "py_threads")
        assert checkers.fds
        assert checkers.py_threads
        assert not checkers.handles
        assert not checkers.c_threads
        assert not checkers.memory
        assert not checkers.gcgarbage

        with pytest.raises(ValueError, match="invalid_checker"):
            Checkers.only("fds", "invalid_checker")

    def test_only_with_all_fields(self):
        # should enable all
        all_fields = Checkers.__annotations__.keys()
        checkers = Checkers.only(*all_fields)
        for f in all_fields:
            assert getattr(checkers, f)

    def test_exclude(self):
        checkers = Checkers.exclude("memory", "fds")
        assert not checkers.memory
        assert not checkers.fds
        assert checkers.handles
        assert checkers.py_threads
        assert checkers.c_threads
        assert checkers.gcgarbage

        with pytest.raises(ValueError, match="not_a_checker"):
            Checkers.exclude("fds", "not_a_checker")

    def test_exclude_with_no_fields(self):
        # should disable nothing, i.e., default True
        checkers = Checkers.exclude()
        for f in Checkers.__annotations__:
            assert getattr(checkers, f)


class TestMemoryLeakTestCaseConfig:

    def test_memory_disabled(self):
        checkers = Checkers.exclude("memory")

        class MyTest(MemoryLeakTestCase):
            pass

        test = MyTest()
        with mock.patch.object(test, "_check_mem", wraps=test._check_mem) as m:
            test.execute(lambda: None, checkers=checkers)
            m.assert_not_called()

    def test_py_threads_disabled(self):
        checkers = Checkers.exclude("py_threads")

        class MyTest(MemoryLeakTestCase):
            pass

        test = MyTest()
        with mock.patch.object(
            threading, "active_count", wraps=threading.active_count
        ) as m:
            test.execute(lambda: None, checkers=checkers)
            m.assert_not_called()


class TestEmitWarnings:
    def setup_method(self):
        psleak._warnings_emitted = False

    def assert_warn_msg(self, msg):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            _emit_warnings()
        assert len(w) == 1
        assert msg in str(w[0].message)

    def test_pythonmalloc_not_set(self, monkeypatch):
        monkeypatch.delenv("PYTHONMALLOC", raising=False)
        self.assert_warn_msg(
            "PYTHONMALLOC=malloc environment variable was not set"
        )

    def test_pythonunbuffered_not_set(self, monkeypatch):
        monkeypatch.delenv("PYTHONUNBUFFERED", raising=False)
        self.assert_warn_msg(
            "PYTHONUNBUFFERED=1 environment variable was not set"
        )

    def test_pytest_xdist_worker(self, monkeypatch):
        monkeypatch.setenv("PYTEST_XDIST_WORKER", "gw0")
        self.assert_warn_msg("pytest-xdist")

    def test_no_heap_info(self):
        with mock.patch.object(psleak.psutil, "heap_info", new=mock.DEFAULT):
            del psleak.psutil.heap_info
            self.assert_warn_msg("heap_info() not available")

    def test_active_threads_warning(self):
        def fun():
            while not stop:
                time.sleep(0.001)

        stop = False
        thread = threading.Thread(target=fun)
        thread.start()
        try:
            self.assert_warn_msg("active Python threads exist")
        finally:
            stop = True
            thread.join()



# Copyright (c) 2025, Giampaolo Rodola. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import threading

import pytest
import test_ext as cext
from psutil import POSIX
from psutil import WINDOWS

from psleak import MemoryLeakError
from psleak import MemoryLeakTestCase
from psleak import UnclosedHeapCreateError
from psleak import UnclosedNativeThreadError

from . import retry_on_failure


class TestMallocWithoutFree(MemoryLeakTestCase):
    """Allocate memory via malloc() and deliberately never call free().
    This must trigger a MemoryLeakError because `heap_used` grows for
    small allocations, and `mmap_used` grows for bigger ones.
    """

    def run_test(self, size, times=None):
        # just malloc(); expect failure
        with pytest.raises(MemoryLeakError):
            self.execute(cext.malloc, size, times=times)

        # malloc() + free(); expect success
        def fun():
            ptr = cext.malloc(size)
            cext.free(ptr)

        self.execute(fun, times=times)

    def test_1b(self):
        self.run_test(1, times=200)

    def test_1k(self):
        self.run_test(1024)

    def test_16k(self):
        self.run_test(1024 * 16)

    @retry_on_failure()
    def test_1M(self):  # noqa: N802
        self.run_test(1024 * 1024, times=30)


# --- posix


@pytest.mark.skipif(not POSIX, reason="POSIX only")
class TestMmapWithoutMunmap(TestMallocWithoutFree):
    """Allocate memory via mmap() and deliberately never call munmap().
    Funnily enough it's not `mmap_used` that grows but VMS.
    """

    def run_test(self, size, times=None):
        # just mmap(); expect failure
        with pytest.raises(MemoryLeakError):
            self.execute(cext.mmap, size, times=times)

        # mmap() + munmap(); expect success
        def fun():
            ptr = cext.mmap(size)
            cext.munmap(ptr, size)

        self.execute(fun, times=times)


# --- windows


@pytest.mark.skipif(not WINDOWS, reason="WINDOWS only")
class TestHeapAllocWithoutHeapFree(TestMallocWithoutFree):
    """Allocate memory via HeapAlloc() and deliberately never call
    HeapFree().
    """

    def run_test(self, size, times=None):
        # just HeapAlloc(); expect failure
        with pytest.raises(MemoryLeakError):
            self.execute(cext.HeapAlloc, size, times=times)

        # HeapAlloc() + HeapFree(); expect success
        def fun():
            ptr = cext.HeapAlloc(size)
            cext.HeapFree(ptr)

        self.execute(fun, times=times)


@pytest.mark.skipif(not WINDOWS, reason="WINDOWS only")
class TestVirtualAllocWithoutVirtualFree(TestMallocWithoutFree):
    """Allocate memory via VirtualAlloc() and deliberately never call
    VirtualFree().
    """

    def run_test(self, size, times=None):
        # just VirtualAlloc(); expect failure
        with pytest.raises(MemoryLeakError):
            self.execute(cext.VirtualAlloc, size, times=times)

        # VirtualAlloc() + VirtualFree(); expect success
        def fun():
            ptr = cext.VirtualAlloc(size)
            cext.VirtualFree(ptr)

        self.execute(fun, times=times)


@pytest.mark.skipif(not WINDOWS, reason="WINDOWS only")
class TestHeapCreateWithoutHeapDestroy(TestMallocWithoutFree):
    """Allocate memory via HeapCreate() and deliberately never call
    HeapDestroy(). Expect UnclosedHeapCreateError to be raised.
    """

    def run_test(self, size, times=None):
        # just HeapCreate(); expect failure
        with pytest.raises(UnclosedHeapCreateError):
            self.execute(cext.HeapCreate, size, 0, times=times)

        # HeapCreate() + HeapDestroy(); expect success
        def fun():
            ptr = cext.HeapCreate(size, 0)
            cext.HeapDestroy(ptr)

        self.execute(fun, times=times)


# --- threads


class TestUnclosedThreads(MemoryLeakTestCase):

    def test_c_thread(self):
        """Create a native C thread and leave it running. Expect
        UnclosedNativeThreadError to be raised.
        """

        def fun():
            nonlocal ptr
            ptr = cext.start_native_thread()
            # Native C threads are supposed to be "hidden" to Python.
            # Make sure they doesn't show up.
            assert threading.active_count() == init_pythread_count
            self.addCleanup(cext.stop_native_thread, ptr)

        init_pythread_count = threading.active_count()
        ptr = None
        with pytest.raises(UnclosedNativeThreadError):
            self.execute(fun)


# --- python idioms


class TestPythonExtensionLeaks(MemoryLeakTestCase):
    """Test typical patterns that lead to a memory leak in C
    extensions, like forgetting Py_DECREF, etc.
    """

    def test_leak_list_small(self):
        # fails without PYTHONMALLOC=malloc
        with pytest.raises(MemoryLeakError):
            self.execute(cext.leak_list, 1)

    def test_leak_list_big(self):
        with pytest.raises(MemoryLeakError):
            self.execute(cext.leak_list, 100)

    def test_leak_long_small(self):
        # fails without PYTHONMALLOC=malloc
        with pytest.raises(MemoryLeakError):
            self.execute(cext.leak_long, 512)

    def test_leak_long_big(self):
        with pytest.raises(MemoryLeakError):
            self.execute(cext.leak_long, 1024)

    def test_leak_tuple_small(self):
        # fails without PYTHONMALLOC=malloc
        with pytest.raises(MemoryLeakError):
            self.execute(cext.leak_tuple, 1)

    def test_leak_tuple_big(self):
        with pytest.raises(MemoryLeakError):
            self.execute(cext.leak_tuple, 100)

    def test_leak_dict(self):
        with pytest.raises(MemoryLeakError):
            self.execute(cext.leak_dict)

    def test_leak_cycle(self):
        with pytest.raises(MemoryLeakError):
            self.execute(cext.leak_cycle)


class TestPymalloc(MemoryLeakTestCase):

    def run_test(self, alloc_fun, free_fun, size, times=None):
        # just allocate; expect failure
        with pytest.raises(MemoryLeakError):
            self.execute(alloc_fun, size, times=times)

        # allocate + free; expect success
        def fun():
            ptr = alloc_fun(size)
            free_fun(ptr)

        self.execute(fun, times=times)

    def test_pymem_malloc_small(self):
        self.run_test(cext.pymem_malloc, cext.pymem_free, 1)

    def test_pymem_malloc_big(self):
        self.run_test(cext.pymem_malloc, cext.pymem_free, 1024)

    def test_pyobject_malloc_small(self):
        self.run_test(cext.pyobject_malloc, cext.pyobject_free, 1)

    def test_pyobject_malloc_big(self):
        self.run_test(cext.pyobject_malloc, cext.pyobject_free, 1024)



# Copyright (c) 2025, Giampaolo Rodola. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""Test memory leak detection heurisics."""

import unittest

import pytest

from psleak import MemoryLeakError
from psleak import MemoryLeakTestCase


class DummyMemLeakTest(MemoryLeakTestCase):
    def __init__(self, diffs_seq):
        super().__init__("runTest")
        self._diffs_seq = iter(diffs_seq)
        self._printed = []

    def _call_ntimes(self, fun, times):
        return next(self._diffs_seq)

    def _log(self, msg, level):
        super()._log(msg, level)
        self._printed.append(msg)

    def printed(self):
        return "".join(self._printed)

    def runs_count(self):
        return self.printed().count("Run # ")

    def call(self, fun):
        return None


def noop():
    pass


class TestMemleakDetectionAlgo(unittest.TestCase):

    def test_increase(self):
        diffs = [
            {"heap": 1024, "uss": 0, "rss": 0, "vms": 0, "mmap": 0},
            {"heap": 2048, "uss": 0, "rss": 0, "vms": 0, "mmap": 0},
        ]
        t = DummyMemLeakTest(diffs)
        with pytest.raises(MemoryLeakError):
            t.execute(noop, retries=len(diffs))

    def test_decrease(self):
        diffs = [
            {"heap": 2048, "uss": 0, "rss": 0, "vms": 0, "mmap": 0},
            {"heap": 1024, "uss": 0, "rss": 0, "vms": 0, "mmap": 0},
        ]
        t = DummyMemLeakTest(diffs)
        t.execute(noop, retries=len(diffs))
        assert "no further growth" in t.printed()

    def test_same(self):
        diffs = [
            {"heap": 1024, "uss": 8192, "rss": 0, "vms": 0, "mmap": 0},
            {"heap": 1024, "uss": 8192, "rss": 0, "vms": 0, "mmap": 0},
            {"heap": 1024, "uss": 8192, "rss": 0, "vms": 0, "mmap": 0},
        ]
        t = DummyMemLeakTest(diffs)
        t.execute(noop, retries=len(diffs))
        assert "no further growth" in t.printed()
        assert t.runs_count() == 2

    # ---

    def test_partial_decrease(self):
        # scenario: heap the same, uss decreased
        diffs = [
            {"heap": 1024, "uss": 20480, "rss": 0, "vms": 0, "mmap": 0},
            {"heap": 1024, "uss": 8192, "rss": 0, "vms": 0, "mmap": 0},
            {"heap": 1024, "uss": 4096, "rss": 0, "vms": 0, "mmap": 0},
        ]
        t = DummyMemLeakTest(diffs)
        t.execute(noop, retries=len(diffs))
        assert "no further growth" in t.printed()
        assert t.runs_count() == 2

    def test_new_metric_appears(self):
        diffs = [
            {"heap": 1024, "uss": 8192, "rss": 0, "vms": 0, "mmap": 0},
            {"heap": 1024, "uss": 8192, "rss": 4096, "vms": 0, "mmap": 0},
        ]
        t = DummyMemLeakTest(diffs)
        with pytest.raises(MemoryLeakError):
            t.execute(noop, retries=len(diffs))

    def test_metric_disappears(self):
        diffs = [
            {"heap": 1024, "uss": 8192, "rss": 4096, "vms": 0, "mmap": 0},
            {"heap": 1024, "uss": 8192, "rss": 0, "vms": 0, "mmap": 0},
        ]
        t = DummyMemLeakTest(diffs)
        t.execute(noop, retries=len(diffs))
        assert "no further growth" in t.printed()



# Copyright (c) 2025, Giampaolo Rodola. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

from psutil import POSIX
from psutil import WINDOWS
from setuptools import Extension
from setuptools import setup

macros = []

if POSIX:
    macros.append(("PSLEAK_POSIX", 1))
if WINDOWS:
    macros.append(("PSLEAK_WINDOWS", 1))

setup(
    name="test_ext",
    ext_modules=[
        Extension(
            "test_ext",
            ["tests/test_ext.c"],
            define_macros=macros,
        )
    ],
)



# Copyright (c) 2025, Giampaolo Rodola. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import threading

import pytest

from psleak import Checkers
from psleak import GCDebugger
from psleak import MemoryLeakTestCase
from psleak import UnclosedPythonThreadError
from psleak import UncollectableGarbageError


class TestPythonThreads(MemoryLeakTestCase):

    def test_it(self):
        """Create a Python thread and leave it running (no join()).
        Expect UnclosedPythonThreadError to be raised.
        """

        def worker():
            done.wait()  # block until signaled

        def fun():
            thread = threading.Thread(target=worker)
            thread.start()
            self.addCleanup(done.set)

        done = threading.Event()
        with pytest.raises(UnclosedPythonThreadError):
            self.execute(fun)


class TestUncollectableGarbage(MemoryLeakTestCase):
    checkers = Checkers.only("gcgarbage")

    def test_simple_cycle(self):
        class Leaky:
            def __init__(self):
                self.ref = None

        def create_cycle():
            a = Leaky()
            b = Leaky()
            a.ref = b
            b.ref = a
            return a, b  # cycle preventing GC from collecting

        with pytest.raises(UncollectableGarbageError):
            self.execute(create_cycle)

    def test_complex_cycle(self):
        class A:
            def __init__(self):
                self.ref = None

        class B:
            def __init__(self):
                self.ref = None

        def create_complex_cycle():
            a = A()
            b = B()
            c = {"x": a, "y": b}
            a.ref = c
            b.ref = c
            return a, b, c

        with pytest.raises(UncollectableGarbageError):
            self.execute(create_complex_cycle)

    def test_self_referencing_object(self):
        class Leaky:  # noqa: B903
            def __init__(self):
                self.ref = self  # self-cycle

        def create_self_cycle():
            a = Leaky()
            return a

        with pytest.raises(UncollectableGarbageError):
            self.execute(create_self_cycle)

    def test_cycle_in_container(self):
        class Node:
            def __init__(self, name):
                self.name = name
                self.children = []

        def create_nested_cycle():
            n1 = Node("n1")
            n2 = Node("n2")
            n1.children.append(n2)
            n2.children.append(n1)  # creates cycle
            return n1, n2

        with pytest.raises(UncollectableGarbageError):
            self.execute(create_nested_cycle)

    def test_ignores_exception_objects(self):
        def create_exception():
            try:
                raise ValueError("test")  # noqa: TRY301
            except ValueError as e:
                err = e  # stored in local variable
            return err

        self.execute(create_exception)


class TestGCDebugger(MemoryLeakTestCase):

    def test_is_transient_ignores(self):
        with GCDebugger() as dbg:
            # scalar objects
            for obj in [1, 2.0, True, "foo", b"bytes", None]:
                assert dbg.is_transient(obj)
            # MainThread
            assert dbg.is_transient(threading.current_thread())
            # exceptions
            assert dbg.is_transient(ValueError())

    def test_nested_containers_with_transient_objects(self):
        t = threading.current_thread()
        nested = [1, (2, [t]), {3, 4}, {"x": t}]

        with GCDebugger() as dbg:
            dbg.after.extend(nested)

        # All objects should be classified as transient because
        # contents are either scalar or transient.
        for obj in nested:
            assert dbg.is_transient(obj)
